@@ -1,37 +1,42 @@
 #!/usr/bin/env python
 # -*- coding: utf-8 -*-
+# see license https://github.com/DerwenAI/kglab#license-and-copyright
 
 ######################################################################
 ## kglab - core classes
 
+from kglab.decorators import multifile
 from kglab.pkg_types import PathLike, IOPathLike, GraphLike, RDF_Node
+from kglab.gpviz import GPViz
 from kglab.util import get_gpu_count
 from kglab.version import _check_version
 _check_version()
 
-import rdflib  # type: ignore
-import rdflib.plugin  # type: ignore
-import rdflib.plugins.parsers.notation3 as rdf_n3  # type: ignore
-rdflib.plugin.register("json-ld", rdflib.plugin.Parser, "rdflib_jsonld.parser", "JsonLDParser")
-rdflib.plugin.register("json-ld", rdflib.plugin.Serializer, "rdflib_jsonld.serializer", "JsonLDSerializer")
+import rdflib  # type: ignore  # pylint: disable=E0401
+import rdflib.plugin  # type: ignore  # pylint: disable=E0401
+import rdflib.plugins.parsers.notation3 as rdf_n3  # type: ignore  # pylint: disable=E0401
+#rdflib.plugin.register("json-ld", rdflib.plugin.Parser, "rdflib_jsonld.parser", "JsonLDParser")
+#rdflib.plugin.register("json-ld", rdflib.plugin.Serializer, "rdflib_jsonld.serializer", "JsonLDSerializer")
 
-from icecream import ic  # type: ignore
-import chocolate  # type: ignore
+from icecream import ic  # type: ignore  # pylint: disable=E0401
+import chocolate  # type: ignore  # pylint: disable=E0401
 import codecs
-import csvwlib  # type: ignore
-import dateutil.parser as dup
+import csvwlib  # type: ignore  # pylint: disable=E0401
+import datetime
+import dateutil.parser as dup  # pylint: disable=E0401
 import io
 import json
-import owlrl  # type: ignore
-import pandas as pd  # type: ignore
+import owlrl  # type: ignore  # pylint: disable=E0401
+import pandas as pd  # type: ignore  # pylint: disable=E0401
 import pathlib
-import pyshacl  # type: ignore
+import pyshacl  # type: ignore  # pylint: disable=E0401
+import pyvis  # type: ignore  # pylint: disable=E0401
 import traceback
 import typing
-import urlpath  # type: ignore
+import urlpath  # type: ignore  # pylint: disable=E0401
 
 if get_gpu_count() > 0:
-    import cudf  # type: ignore # pylint: disable=E0401
+    import cudf  # type: ignore  # pylint: disable=E0401
 
 
 class KnowledgeGraph:
@@ -220,6 +225,11 @@ a `dict` describing the namespaces in this RDF graph
             for prefix, ns in self._ns.items()
         }
 
+        nm = self._g.namespace_manager
+
+        for prefix, uri in nm.namespaces():
+            ns_dict[prefix] = str(uri)
+
         return ns_dict
 
 
@@ -239,7 +249,7 @@ a [`pandas.DataFrame`](https://pandas.pydata.org/pandas-docs/stable/reference/ap
                 col_names[0]: prefix,
                 col_names[1]: str(ns),
             }
-            for prefix, ns in self._ns.items()
+            for prefix, ns in self.get_ns_dict().items()
         ]
 
         if self.use_gpus:
@@ -271,13 +281,13 @@ context needed for JSON-LD serialization
 
     def encode_date (
         self,
-        datetime: str,
+        dt: str,
         tzinfos: dict,
         ) -> rdflib.Literal:
         """
 Helper method to ensure that an input `datetime` value has a timezone that can be interpreted by [`rdflib.XSD.dateTime`](https://www.w3.org/TR/xmlschema-2/#dateTime).
 
-    datetime:
+    dt:
 input datetime as a string
 
     tzinfos:
@@ -287,7 +297,7 @@ timezones as a dict, used by
     returns:
 [`rdflib.Literal`](https://rdflib.readthedocs.io/en/stable/rdf_terms.html#literals) formatted as an XML Schema 2 `dateTime` value
         """
-        date_tz = dup.parse(datetime, tzinfos=tzinfos)
+        date_tz = dup.parse(dt, tzinfos=tzinfos)
         return rdflib.Literal(date_tz, datatype=self.get_ns("xsd").dateTime)
 
 
@@ -457,6 +467,7 @@ a string as a file name or URL to a file reference
         return filename
 
 
+    @multifile()
     def load_rdf (
         self,
         path: IOPathLike,
@@ -473,7 +484,7 @@ Throws `TypeError` whenever a format parser plugin encounters a syntax error.
 Note: this adds relations to an RDF graph, although it does not overwrite the existing RDF graph.
 
     path:
-must be a file name (str) or a path object (not a URL) to a local file reference; or a [*readable, file-like object*](https://docs.python.org/3/glossary.html#term-file-object)
+must be a file name (str) to a local file reference – possibly a glob with a wildcard; or a path object (not a URL) to a local file reference; or a [*readable, file-like object*](https://docs.python.org/3/glossary.html#term-file-object); otherwise this throws a `TypeError` exception
 
     format:
 serialization format, defaults to Turtle triples; see `_RDF_FORMAT` for a list of default formats, which can be extended with plugins – excluding the `"json-ld"` format; otherwise this throws a `TypeError` exception
@@ -582,9 +593,9 @@ serialization format, which defaults to Turtle triples; see `_RDF_FORMAT` for a
 optional base set for the graph
 
     encoding:
-text encoding value, defaults to `"utf-8"`, must be in the [Python codec registry](https://docs.python.org/3/library/codecs.html#codecs.CodecInfo); otherwise this throws a `LookupError` exception
+optional text encoding value, defaults to `"utf-8"`, must be in the [Python codec registry](https://docs.python.org/3/library/codecs.html#codecs.CodecInfo); otherwise this throws a `LookupError` exception
         """
-        # error checking for the `format` paramter
+        # error checking for the `format` parameter
         if format == "json-ld":
             raise TypeError("Use the save_jsonld() method instead")
 
@@ -597,7 +608,7 @@ text encoding value, defaults to `"utf-8"`, must be in the [Python codec registr
         if not base and self.base_uri:
             base = self.base_uri
 
-        # error checking for a file-like object `path` paramter
+        # error checking for a file-like object `path` parameter
         if hasattr(path, "write"):
             if hasattr(path, "encoding"):
                 raise TypeError(self._ERROR_PATH)
@@ -642,7 +653,7 @@ serialization format, which defaults to Turtle triples; see `_RDF_FORMAT` for a
 optional base set for the graph
 
     encoding:
-text encoding value, defaults to `"utf-8"`, must be in the [Python codec registry](https://docs.python.org/3/library/codecs.html#codecs.CodecInfo); otherwise this throws a `LookupError` exception
+optional text encoding value, defaults to `"utf-8"`, must be in the [Python codec registry](https://docs.python.org/3/library/codecs.html#codecs.CodecInfo); otherwise this throws a `LookupError` exception
 
     returns:
 text representing the RDF graph
@@ -666,6 +677,7 @@ text representing the RDF graph
         ).decode(encoding)
 
 
+    @multifile()
     def load_jsonld (
         self,
         path: IOPathLike,
@@ -680,23 +692,26 @@ This traps some edge cases for the several source-ish parameters in RDFlib which
 Note: this adds relations to an RDF graph, it does not overwrite the existing RDF graph.
 
     path:
-must be a file name (str) or a path object (not a URL) to a local file reference; or a [*readable, file-like object*](https://docs.python.org/3/glossary.html#term-file-object); otherwise this throws a `TypeError` exception
+must be a file name (str) to a local file reference – possibly a glob with a wildcard; or a path object (not a URL) to a local file reference; or a [*readable, file-like object*](https://docs.python.org/3/glossary.html#term-file-object)
 
     encoding:
-text encoding value, which defaults to `"utf-8"`; must be in the [Python codec registry](https://docs.python.org/3/library/codecs.html#codecs.CodecInfo); otherwise this throws a `LookupError` exception
+optional text encoding value, which defaults to `"utf-8"`; must be in the [Python codec registry](https://docs.python.org/3/library/codecs.html#codecs.CodecInfo); otherwise this throws a `LookupError` exception
 
     returns:
 this `KnowledgeGraph` object – used for method chaining
         """
-        # error checking for a file-like object `path` paramter
-        if hasattr(path, "read"):
-            f = path
-        else:
-            f = open(path, "r", encoding=encoding) # type: ignore
-
         # error checking for the `encoding` parameter
         self._check_encoding(encoding)
 
+        # error checking for a file-like object `path` parameter
+        if hasattr(path, "read"):
+            f = path
+        else:
+            f = open(path, "r", encoding=encoding)  # type: ignore
+
+        # load JSON from file (to verify format and trap exceptions at
+        # this level) then dump to string – which is expected by the
+        # JSON-LD plugin for RDFlib
         self._g.parse(
             data=json.dumps(json.load(f)),  # type: ignore
             format="json-ld",
@@ -722,9 +737,9 @@ This traps some edge cases for the `destination` parameter in RDFlib which had b
 must be a file name (str) or a path object (not a URL) to a local file reference; or a [*writable, bytes-like object*](https://docs.python.org/3/glossary.html#term-bytes-like-object); otherwise this throws a `TypeError` exception
 
     encoding:
-text encoding value, which defaults to `"utf-8"`; must be in the [Python codec registry](https://docs.python.org/3/library/codecs.html#codecs.CodecInfo); otherwise this throws a `LookupError` exception
+optional text encoding value, which defaults to `"utf-8"`; must be in the [Python codec registry](https://docs.python.org/3/library/codecs.html#codecs.CodecInfo); otherwise this throws a `LookupError` exception
         """
-        # error checking for a file-like object `path` paramter
+        # error checking for a file-like object `path` parameter
         if hasattr(path, "write"):
             if hasattr(path, "encoding"):
                 raise TypeError(self._ERROR_PATH)
@@ -747,33 +762,13 @@ text encoding value, which defaults to `"utf-8"`; must be in the [Python codec r
         )
 
 
-    def load_csv (
-        self,
-        url: str,
-        ) -> "KnowledgeGraph":
-        """
-Wrapper for [`csvwlib`](https://github.com/DerwenAI/csvwlib) which parses a CSV file from the `path` source, then converts to RDF and merges into this RDF graph.
-
-    url:
-must be a URL represented as a string
-
-    returns:
-this `KnowledgeGraph` object – used for method chaining
-        """
-        new_rdf = csvwlib.CSVWConverter.to_rdf(
-            url,
-            mode="minimal",
-            format="ttl",
-        )
-        return self.load_rdf_text(new_rdf)
-
-
     _PARQUET_COL_NAMES: typing.List[str] = [
         "subject",
         "predicate",
         "object"
     ]
 
+    @multifile()
     def load_parquet (
         self,
         path: IOPathLike,
@@ -788,7 +783,7 @@ To prepare for upcoming **kglab** features, **this is the preferred method for d
 Note: this adds relations to an RDF graph, it does not overwrite the existing RDF graph.
 
     path:
-must be a file name (str), path object to a local file reference, or a [*readable, file-like object*](https://docs.python.org/3/glossary.html#term-file-object); a string could be a URL; valid URL schemes include `https`, `http`, `ftp`, `s3`, `gs`, `file`; a file URL can also be a path to a directory that contains multiple partitioned files, including a bucket in cloud storage – based on [`fsspec`](https://github.com/intake/filesystem_spec)
+must be a file name (str) to a local file reference – possibly a glob with a wildcard; or a path object (not a URL) to a local file reference; or a [*readable, file-like object*](https://docs.python.org/3/glossary.html#term-file-object); a string could be a URL; valid URL schemes include `https`, `http`, `ftp`, `s3`, `gs`, `file`; a file URL can also be a path to a directory that contains multiple partitioned files, including a bucket in cloud storage – based on [`fsspec`](https://github.com/intake/filesystem_spec)
 
     returns:
 this `KnowledgeGraph` object – used for method chaining
@@ -858,6 +853,126 @@ extra options parsed by [`fsspec`](https://github.com/intake/filesystem_spec) fo
         )
 
 
+    def load_csv (
+        self,
+        url: str,
+        ) -> "KnowledgeGraph":
+        """
+Wrapper for [`csvwlib`](https://github.com/DerwenAI/csvwlib) which parses a CSV file from the `path` source, then converts to RDF and merges into this RDF graph.
+
+    url:
+must be a URL represented as a string
+
+    returns:
+this `KnowledgeGraph` object – used for method chaining
+        """
+        new_rdf = csvwlib.CSVWConverter.to_rdf(
+            url,
+            mode="minimal",
+            format="ttl",
+        )
+        return self.load_rdf_text(new_rdf)
+
+
+    def _walk_roam_graph (
+        self,
+        obj: dict,
+        seen_uid: typing.Set[str] = None,
+        ) -> str:
+        """
+Semiprivate method to traverse the Roam Research exported graph recursively, converting its objects into RDF representation.
+
+    obj:
+object to parse
+
+    returns:
+The `uid` identifier for the parsed object
+        """
+        if not seen_uid:
+            seen_uid = set()
+
+        roam_ns = str(self.get_ns("roam"))
+        uid = obj["uid"]
+
+        if uid not in seen_uid:
+            seen_uid.add(uid)
+
+            # create a node for this object
+            node = rdflib.URIRef(roam_ns + uid)
+            self.add(node, self.get_ns("rdf").type, self.get_ns("dct").Text)
+
+            # represent title (complex object) or string description (simple object)
+            if "title" in obj:
+                descrip = obj["title"]
+            elif "string" in obj:
+                descrip = obj["string"]
+
+            self.add(node, self.get_ns("skos").definition, rdflib.Literal(descrip))
+
+            # represent the user who created/edited this object
+            user_uid = obj[":edit/user"][":user/uid"]
+            self.add(node, self.get_ns("dct").Creator, rdflib.URIRef(roam_ns + user_uid))
+
+            # convert millisec timestamp to Unix epoch times (UTC) to datetime
+            dt = datetime.datetime.utcfromtimestamp(round(obj["edit-time"] / 1000.0))
+            self.add(node, self.get_ns("dct").Date, rdflib.Literal(dt.isoformat(), datatype=rdflib.XSD.dateTime))
+
+            if "children" in obj:
+                for child_obj in obj["children"]:
+                    child_uid = self._walk_roam_graph(child_obj, seen_uid)
+                    self.add(node, self.get_ns("dct").references, rdflib.URIRef(roam_ns + child_uid))
+
+        return uid
+
+
+    @multifile()
+    def import_roam (
+        self,
+        path: IOPathLike,
+        *,
+        encoding: str = "utf-8",
+        ) -> typing.List[str]:
+        """
+Import a graph in JSON that has been exported from the [Roam Research](https://roamresearch.com/) note-taking tool, then convert its objects and attributes into RDF representation.
+
+For more details about the exported data from Roam Research, see:
+
+  * <https://roamstack.com/roam-data-outside-roam/>
+  * <https://nesslabs.com/roam-research-input-output>
+  * <https://davidbieber.com/snippets/2020-04-25-roam-json-export/>
+
+Note: this adds relations to an RDF graph, it does not overwrite the existing RDF graph.
+
+    path:
+must be a file name (str) to a local file reference – possibly a glob with a wildcard; or a path object (not a URL) to a local file reference; or a [*readable, file-like object*](https://docs.python.org/3/glossary.html#term-file-object)
+
+    encoding:
+optional text encoding value, which defaults to `"utf-8"`; must be in the [Python codec registry](https://docs.python.org/3/library/codecs.html#codecs.CodecInfo); otherwise this throws a `LookupError` exception
+
+    returns:
+a list of identifiers for the top-level nodes added from the Roam Research graph
+        """
+        # error checking for the `encoding` parameter
+        self._check_encoding(encoding)
+
+        # error checking for a file-like object `path` parameter
+        if hasattr(path, "read"):
+            f = path
+        else:
+            f = open(path, "r", encoding=encoding)  # type: ignore
+
+        # add a `roam:` prefix for a pseudo-namespace to use here,
+        # which applications may need to parameterize later?
+        self.add_ns("roam", "https://roamresearch.com/ns/")
+
+        uid_list: typing.List[str] = [
+            self._walk_roam_graph(obj)
+            for obj in json.load(f)  # type: ignore
+            ]
+
+        return uid_list
+
+
     def n3fy (
         self,
         node: RDF_Node,
@@ -877,11 +992,11 @@ flag to force instances of [`rdflib.term.Literal`](https://rdflib.readthedocs.io
 text (or Python objects) for the serialized node
         """
         if pythonify and isinstance(node, rdflib.term.Literal):
-            ser = node.toPython()
+            serialized = node.toPython()
         else:
-            ser = node.n3(self._g.namespace_manager)
+            serialized = node.n3(self._g.namespace_manager)
 
-        return ser
+        return serialized
 
 
     def n3fy_row (
@@ -985,6 +1100,27 @@ the query result set represented as a [`pandas.DataFrame`](https://pandas.pydata
         return df
 
 
+    def visualize_query (
+        self,
+        sparql: str,
+        *,
+        notebook: bool = False,
+        ) -> pyvis.network.Network:
+        """
+Visualize the given SPARQL query as a [`pyvis.network.Network`](https://pyvis.readthedocs.io/en/latest/documentation.html#pyvis.network.Network)
+
+    sparql:
+input SPARQL query to be visualized
+
+    notebook:
+optional boolean flag, whether to initialize the PyVis graph to render within a notebook; defaults to `False`
+
+    returns:
+PyVis network object, to be rendered
+        """
+        return GPViz(sparql, self._ns).visualize_query(notebook=notebook)
+
+
     ######################################################################
     ## SHACL validation
 
@@ -998,7 +1134,7 @@ the query result set represented as a [`pandas.DataFrame`](https://pandas.pydata
         advanced: typing.Optional[bool] = False,
         inference: typing.Optional[str] = None,
         inplace:typing.Optional[bool] = True,
-        abort_on_error: typing.Optional[bool] = None,
+        abort_on_first: typing.Optional[bool] = None,
         **kwargs: typing.Any,
         ) -> typing.Tuple[bool, "KnowledgeGraph", str]:
         """
@@ -1025,7 +1161,7 @@ prior to validation, run OWL2 RL profile-based expansion of the RDF graph based
     inplace:
 when enabled, do not clone the RDF graph prior to inference/expansion, just manipulate it in-place
 
-    abort_on_error:
+    abort_on_first:
 abort validation on the first error
 
     returns:
@@ -1040,7 +1176,7 @@ a tuple of `conforms` (RDF graph passes the validation rules) + `report_graph` (
             advanced=advanced,
             inference=inference,
             inplace=inplace,
-            abort_on_error=abort_on_error,
+            abort_on_first=abort_on_first,
             serialize_report_graph="ttl",
             **chocolate.filter_args(kwargs, pyshacl.validate),
             )
